RPL ( Ensure RPLCOMP and the syntax highlighter are in RPL mode )
( C:\Dokumente und Einstellungen\ag\Eigene Dateien\HP\Projects\GRW\graph.s, part of the GRW.hpp project, created by <> on 15.12.2001 )

INCLUDE GRW.H

* This module implements basic graph operations for internal use in the GRW
* library only. See graphops.s for user-level operations. Graphs are encoded
* in adjacency list form as a list of lists, with optional node embeddings and
* node and edge labels included. Node indices must be specified as untagged
* real values, and must point to a valid entry in the graph.

* Pseudo BNF syntax of the graph data structure:

* graph ::= { node-info ... }
* node-info ::= { [ point ] [ label ... ] [ edge-info ... ] }
* edge-info ::= node | { node [ label ... ] }
* label ::= tagged
* node ::= untagged real
* point ::= untagged complex

* value ops ******************************************************************

* GRNodeList: ( graph -> { node ... } )
* returns the list of all nodes (1..size(graph))

NULLNAME GRNodeList
::
  DUPNULLCOMP? ?SEMI
  LENCOMP DUP 1LAMBIND
  #1+_ONE_DO (DO)
    INDEX@ UNCOERCE
  LOOP
  1GETABND {}N
;

* GRGetNode: ( graph node -> node-info )
* returns the node-info entry of the given node

NULLNAME GRGetNode
::
  COERCE NTHCOMPDROP
;

* GRSplitNode: ( node-info -> point|NOVAL { label ... } { edge-info ... } )
* split node-info into its constituents

NULLNAME GRSplitNode
::
  DUP CARCOMP DUPTYPECMP? ITE :: SWAP CDRCOMP ; :: DROP ' xNOVAL SWAP ;
  ' TYPETAGGED? ListSplit
;

* GRNodePoint: ( node-info -> point | NOVAL )
* returns the embedding of a node if it has one, NOVAL otherwise

NULLNAME GRNodePoint
::
  CARCOMP DUPTYPECMP? ?SKIP :: DROP ' xNOVAL ;
;

* GRNodeEdges: ( node-info -> { edge-info ... } )
* returns the list of all edges of a node

NULLNAME GRNodeEdges
::
  GRSplitNode UNROT 2DROP
;

* GRNodeAdj: ( node-info -> { node ... } )
* returns the list of all adjacent nodes (like GRNodeEdges, but with the edge
* labels removed)

NULLNAME GRNodeAdj
::
  GRNodeEdges ' GREdgeTarget ListMap
;

* GRNodeLabels: ( node-info -> { label ... } )
* returns the list of all labels associated with a node

NULLNAME GRNodeLabels
::
  GRSplitNode DROP SWAPDROP
;

* GRNodeLabel: ( node-info tag -> label | NOVAL )
* returns the first node label value (untagged) for the given tag (specified
* as a symbol or string); NOVAL if a value for the given tag does not exist

NULLNAME GRNodeLabel
::
  ( convert tag to string )
  DUPTYPEIDNT? OVER TYPELAM? OR IT ID>$
  ( extract labels from node info )
  SWAP GRNodeLabels DUP ROT
  ( search for matching tags )
  ' ::
    SWAP DUPTYPETAG? ITE ::
      ' xOBJ> EvalNoCK SWAPDROP EQUAL
    ;
    ::
      # 203 ERROROUT
    ;
  ;
  POSCOMP
  DUP #0= case2drop :: ' xNOVAL ;
  NTHCOMPDROP STRIPTAGS
;

* GRSplitEdge: ( edge-info -> node { label ... } )
* splits edge-info into its constituents

NULLNAME GRSplitEdge
::
  DUPTYPELIST? ITE :: DUP CARCOMP SWAP CDRCOMP ; NULL{}
;

* GREdgeTarget: ( edge-info -> node )
* returns the target node of the edge

NULLNAME GREdgeTarget
::
  DUPTYPELIST? IT CARCOMP
;

* GREdgeLabels: ( edge-info -> { label ... } )
* returns the list of all labels associated with an edge

NULLNAME GREdgeLabels
::
  DUPTYPELIST? ITE CDRCOMP :: DROP NULL{} ;
;

* GREdgeLabel: ( edge-info tag -> label | NOVAL )
* returns the first edge label value (untagged) for the given tag (specified
* as a symbol or string); NOVAL if a value for the given tag does not exist

NULLNAME GREdgeLabel
::
  ( convert tag to string )
  DUPTYPEIDNT? OVER TYPELAM? OR IT ID>$
  ( extract labels from edge info )
  SWAP GREdgeLabels DUP ROT
  ( search for matching tags )
  ' ::
    SWAP DUPTYPETAG? ITE ::
      ' xOBJ> EvalNoCK SWAPDROP EQUAL
    ;
    ::
      # 203 ERROROUT
    ;
  ;
  POSCOMP
  DUP #0= case2drop xNOVAL
  NTHCOMPDROP STRIPTAGS
;

* modification ops ***********************************************************

* GRSetNode: ( graph node node-info -> graph' )
* updates the node info of an existing node

NULLNAME GRSetNode
::
  SWAP COERCE ROT PUTLIST
;

* GRMakeNode: ( point|NOVAL { label ... } { edge-info ... } -> node-info )
* create a node-info structure from its constituents

NULLNAME GRMakeNode
::
  &COMP SWAP DUP ' xNOVAL EQUAL ITE DROP >HCOMP
;

* GRSetNodePoint: ( node-info point -> node-info' )
* updates the embedding of a node

NULLNAME GRSetNodePoint
::
  SWAP GRSplitNode ROTDROP GRMakeNode
;

* GRSetNodeEdges: ( node-info { edge-info ... } -> node-info' )
* updates the edges of a node

NULLNAME GRSetNodeEdges
::
  SWAP GRSplitNode DROPROT GRMakeNode
;

* GRSetNodeLabels: ( node-info { label ... ) -> node-info' )
* updates the labels of a node

NULLNAME GRSetNodeLabels
::
  SWAP GRSplitNode SWAPDROP ROTSWAP GRMakeNode
;

* GRSetNodeLabel: ( node-info tag val -> node-info' )
* updates or adds the given node label

NULLNAME GRSetNodeLabel
::
  ( convert tag to string )
  SWAP DUPTYPEIDNT? OVER TYPELAM? OR IT ID>$
  ( store label and tag )
  { NULLLAM NULLLAM } BIND ( 2LAM = val, 1LAM = tag )
  ( locate tag in label list )
  GRSplitNode SWAP DUP 1GETLAM
  ' ::
    SWAP DUPTYPETAG? ITE ::
      ' xOBJ> EvalNoCK SWAPDROP EQUAL
    ;
    FALSE
  ;
  POSCOMP
  ( if present, store in existing entry, otherwise append new entry )
  2GETLAM 1GETLAM USER$>TAG
  SWAP DUP #0= ITE :: DROP >TCOMP ; :: ROT PUTLIST ;
  ( rebuild the node info structure )
  SWAP GRMakeNode
  ABND
;

* GRDelNodeLabel: ( node-info tag -> node-info' )
* delete all node labels for the given tag (specified as a symbol or string)

NULLNAME GRDelNodeLabel
::
  ( convert tag to string )
  DUPTYPEIDNT? OVER TYPELAM? OR IT ID>$ 1LAMBIND
  ( extract labels from node info )
  GRSplitNode SWAP
  ( stack now: point edges labels )
  ( filter out matching tags )
  ' ::
    DUPTYPETAG? IT ::
      DUP ' xOBJ> EvalNoCK SWAPDROP 5GETLAM EQUAL IT DROP
    ;
  ; ListMap
  ( stack now: point edges labels' )
  ( update label list )
  SWAP GRMakeNode
  ABND
;

* GRMakeEdge: ( node { label ... } -> edge-info )
* create an edge structure from its constituents

NULLNAME GRMakeEdge
::
  DUPNULLCOMP? ITE DROP :: SWAP >HCOMP ;
;

* GRSetEdgeTarget: ( edge-info node -> edge-info' )
* updates the target node of an edge

NULLNAME GRSetEdgeTarget
::
  SWAP GRSplitEdge SWAPDROP GRMakeEdge
;

* GRSetEdgeLabels: ( edge-info { label ... } -> edge-info' )
* updates the labels of an edge

NULLNAME GRSetEdgeLabels
::
  SWAP GRSplitEdge DROPSWAP GRMakeEdge
;

* GRSetEdgeLabel: ( edge-info tag val -> edge-info' )
* updates or adds the given edge label

NULLNAME GRSetEdgeLabel
::
  ( convert tag to string )
  SWAP DUPTYPEIDNT? OVER TYPELAM? OR IT ID>$
  ( store label and tag )
  { NULLLAM NULLLAM } BIND ( 2LAM = val, 1LAM = tag )
  ( locate tag in label list )
  GRSplitEdge DUP 1GETLAM
  ' ::
    SWAP DUPTYPETAG? ITE ::
      ' xOBJ> EvalNoCK SWAPDROP EQUAL
    ;
    FALSE
  ;
  POSCOMP
  ( if present, store in existing entry, otherwise append new entry )
  2GETLAM 1GETLAM USER$>TAG
  SWAP DUP #0= ITE :: DROP >TCOMP ; :: ROT PUTLIST ;
  ( rebuild the edge info structure )
  GRMakeEdge
  ABND
;

* GRDelEdgeLabel: ( edge-info tag -> edge-info' )
* delete all edge labels for the given tag (specified as a symbol or string)

NULLNAME GRDelEdgeLabel
::
  ( convert tag to string )
  DUPTYPEIDNT? OVER TYPELAM? OR IT ID>$ 1LAMBIND
  ( extract labels from edge info )
  GRSplitEdge
  ( stack now: target labels )
  ( filter out matching tags )
  ' ::
    DUPTYPETAG? IT ::
      DUP ' xOBJ> EvalNoCK SWAPDROP 5GETLAM EQUAL IT DROP
    ;
  ; ListMap
  ( stack now: target labels' )
  ( update label list )
  GRMakeEdge
  ABND
;

* graph construction *********************************************************

* GRMakeGraph: ( #n -> graph )
* create an edgeless graph with n nodes

NULLNAME GRMakeGraph
::
  DUP#0=case :: DROP NULL{} ;
  1LAMBIND 1GETLAM
  #1+_ONE_DO (DO)
    NULL{}
  LOOP
  1GETABND {}N
;

* GRAddNode: ( graph node-info -> graph' )
* adds a new node, which becomes the last node of the graph

NULLNAME GRAddNode
::
  >TCOMP
;

* GRDelNode: ( graph node -> graph' )
* deletes a node and all incident edges; updates node indices accordingly

NULLNAME GRDelNode
::
  1LAMBIND
  ( remove node )
  1GETLAM COERCE ListSplitAt CDRCOMP &COMP
  ( for all nodes )
  ' ::
    GRSplitNode
    ( for all edges of current node )
    ' ::
      DUP GRSplitEdge
      SWAP
      ( edge-info labels node )
      DUP 9GETLAM %= case
      ( skip incident edge )
      3DROP
      DUP 9GETLAM %> case
      ( update edge )
      :: %1- SWAP GRMakeEdge SWAPDROP ;
      ( default case: keep existing edge )
      2DROP
    ; ListMap
    GRMakeNode
  ; ListMap
  ABND
;

* GRAddEdge: ( graph node node { label ... } -> graph' )
* adds a new edge from the given source to the given target node

NULLNAME GRAddEdge
::
  GRMakeEdge UNROT COERCE 2DUP NTHCOMPDROP 4ROLL >TCOMP SWAPROT PUTLIST
;

* GRAddEdge2:  ( graph node node { label ... } -> graph' )
* adds an edge and its reversal

NULLNAME GRAddEdge2
::
  UNROT 2DUP %= ITE ::
    ROT GRAddEdge
  ;
  ::
    ROT 4ROLL 4PICK 4PICK 4PICK GRAddEdge 4UNROLL UNROT SWAPROT GRAddEdge
  ;
;

* GRDelEdge: ( graph node node -> graph' )
* deletes the first edge going from the given source to the given target node

NULLNAME GRDelEdge
::
  1LAMBIND
  2DUP COERCE NTHCOMPDROP GRSplitNode
  ( graph source point labels edges )
  DUP 1GETABND ' :: SWAP GREdgeTarget %= ; POSCOMP
  DUP #0= ITE 5DROP
  :: ListSplitAt CDRCOMP &COMP GRMakeNode GRSetNode ;
;

* GRDelEdge2: ( graph node node -> graph' )
* deletes an edge and its reversal

NULLNAME GRDelEdge2
::
  2DUP %= ITE GRDelEdge ::
    ROT 3PICK 3PICK GRDelEdge UNROT SWAP GRDelEdge
  ;
;

* graph manipulation ********************************************************

* GREmbed: graph %xmin %xmax %ymin %ymax -> graph'
* embed a graph; adds random coordinates in the specified range to unembedded
* nodes

NULLNAME GREmbed
::
  NULLLAM FOUR NDUPN DOBIND
  ' ::
    GRSplitNode ROT DUPTYPECMP?
    ?SKIP ::
      DROP
      8GETLAM 7GETLAM OVER %- %RAN %* %+
      6GETLAM 5GETLAM OVER %- %RAN %* %+
      %>C%
    ;
    UNROT GRMakeNode
  ; ListMap
  ABND
;

* GRSub: ( graph { node ... } -> graph' )
* construct the subgraph induced by the given nodes, in the given order
* precond: entries in the node list must be pairwise distinct

NULLNAME GRSub
::
  ( construct map from old to new node numbers )
  OVER LENCOMP { NULLLAM NULLLAM } BIND
  toLEN_DO (DO)
    2GETLAM INDEX@ UNCOERCE ' %= POSCOMP UNCOERCE
  LOOP
  1GETLAM {}N 2GETLAM ABND
  ( graph map nodes )
  ( copy subgraph )
  UNROT { NULLLAM NULLLAM } BIND
  ' ::
    6GETLAM SWAP GRGetNode GRSplitNode
    ' ::
      GRSplitEdge SWAP COERCE
      9GETLAM SWAP NTHCOMPDROP
      DUP %0= case2DROP
      SWAP GRMakeEdge
    ; ListMap
    GRMakeNode
  ; ListMap
  ABND
;

* graph parser ***************************************************************

* GRParseGraph: ( graph -> T|F )
* check the syntax of a graph object

NULLNAME GRParseGraph
::
  DUPTYPELIST? ITE ::
    ( record graph size, used to check edge targets )
    DUPLENCOMP UNCOERCE SWAP
    ' :: OVERSWAP GRParseNode ; ListDoWhile
    SWAPDROP
  ;
  DROPFALSE
;

* GRParseNode: ( %nodes node-info -> T|F )
* parses a node-info structure

NULLNAME GRParseNode
::
  DUPTYPELIST? ITE ::
    ( no need to check point and labels, these will always be legal )
    GRNodeEdges GRParseEdges
  ;
  2DROPFALSE
;

* GRParseEdges: ( %nodes { edge-info ... } -> T|F )
* parses an edge-info list

NULLNAME GRParseEdges
::
  DUPTYPELIST? ITE ::
    ' :: OVERSWAP GRParseEdge ; ListDoWhile
    SWAPDROP
  ;
  2DROPFALSE
;

* GRParseEdge: ( %nodes edge-info -> T|F )
* parses an edge-info structure

NULLNAME GRParseEdge
::
  DUPTYPELIST? case ::
    GRSplitEdge
    UNROT GRParseTarget NOTcase DROPFALSE
    GRParseLabels
  ;
  DUPTYPEREAL? case GRParseTarget
  2DROPFALSE
;

* GRParseTarget: ( %nodes node -> T|F )
* parses the target node value

NULLNAME GRParseTarget
::
  DUPTYPEREAL? ITE ::
    DUP %0> NOTcase 2DROPFALSE
    %< caseFALSE
    TRUE
  ;
  2DROPFALSE
;

* GRParseLabels: ( { label ... } -> T|F )
* parses a label list

NULLNAME GRParseLabels
::
  DUPTYPELIST? ITE ::
    ' TYPETAGGED? ListDoWhile
  ;
  DROPFALSE
;

* GRParseNodeList: ( %nodes { node ... } -> T|F )
* parses a node list

NULLNAME GRParseNodeList
::
  DUPTYPELIST? ITE ::
    ' ::
      DUPTYPEREAL? NOTcase DROPFALSE
      DUP %0> NOTcase DROPFALSE
      OVER %<=
    ; ListDoWhile
    SWAPDROP
  ;
  2DROPFALSE
;
