RPL ( Ensure RPLCOMP and the syntax highlighter are in RPL mode )
( D:\Dokumente und Einstellungen\Administrator\Eigene Dateien\HP\Projects\GRW\graphops.s, part of the GRW.hpp project, created by <> on 18.12.2001 )

INCLUDE GRW.H

* This module implements user-level graph operations for interactive
* manipulation of a graph on the stack and for RPL programming. They are based
* on the internal graph operations in graph.s, but check the types of
* arguments being operated upon, to prevent memory corruption and other nasty
* events.

* GRAPH SYNTAX: Graphs are encoded in adjacency list form as a list of lists,
* with optional node embeddings and node and edge labels included. As a
* special case, the empty graph is represented as the empty list. Operations
* are provided for parsing the syntax of graph objects, to create and delete
* nodes and edges, and to access and change the adjacency and label
* information.

* Pseudo BNF syntax:

* graph ::= { node-info ... }
* node-info ::= { [ point ] [ label ... ] [ edge-info ... ] }
* edge-info ::= node | { node [ label ... ] }
* label ::= tagged
* node ::= untagged real
* point ::= untagged complex

* Note that in order to distinguish between labels and other values in the
* adjacency lists, the former must always be entered as tagged values. Node
* indices must be specified as reals, and must point to a valid node entry in
* the graph. These conditions are all enforced when a graph is built using the
* operations provided by this module. The syntax of a graph object entered
* directly by the user can be checked with the GRAPH? predicate.

* The adjacency list structure allows both a compact representation and fast
* access to the outgoing edges of each node (ingoing edges can be accessed
* efficiently using the graph's reversal). It also allows you to represent
* multigraphs, i.e., graphs with multiple edges between the same pair of
* nodes. No special representation is provided for undirected graphs, but they
* can be easily be encoded as bidirected graphs, i.e., graphs with a symmetric
* adjacency relation.

NULLNAME CKNode ( %nodes node -> T|F )
::
  DUPTYPEREAL? NOTcase 2DROPFALSE
  DUP %0> NOTcase 2DROPFALSE
  %>=
;

NULLNAME GRAddEdgeCK
::
  GRMakeEdge UNROT COERCE 2DUP NTHCOMPDROP
  DUPTYPELIST? NOTcase :: # 203 ERROROUT ;
  4ROLL >TCOMP SWAPROT PUTLIST
;

NULLNAME GRAddEdge2CK
::
  UNROT 2DUP %= ITE ::
    ROT GRAddEdgeCK
  ;
  ::
    ROT 4ROLL 4PICK 4PICK 4PICK GRAddEdgeCK 4UNROLL UNROT SWAPROT GRAddEdgeCK
  ;
;

NULLNAME GRDelEdgeCK
::
  1LAMBIND
  2DUP COERCE NTHCOMPDROP
  DUPTYPELIST? NOTcase :: # 203 ERROROUT ;
  GRSplitNode
  ( graph source point labels edges )
  DUP 1GETABND ' ::
    SWAP GREdgeTarget
    DUPTYPEREAL? ITE %= FALSE
  ; POSCOMP
  DUP #0= ITE 5DROP
  ::
    ListSplitAt CDRCOMP &COMP GRMakeNode GRSetNode
  ;
;

NULLNAME GRDelEdge2CK
::
  2DUP %= ITE GRDelEdgeCK
  ::
    ROT 3PICK 3PICK GRDelEdgeCK UNROTSWAP GRDelEdgeCK
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME GRAPH?
::
  CK1
  GRParseGraph COERCEFLAG
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME NODE?
::
  CK1
  %MAXREAL SWAP GRParseNode COERCEFLAG
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME EDGE?
::
  CK1
  %MAXREAL SWAP GRParseEdge COERCEFLAG
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME ADJ?
::
  CK2&Dispatch
  # 51 ::
    SWAP GRNodeAdj matchob? ITE %1 :: DROP %0 ;
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME NGRAPH
::
  CK1&Dispatch
  #1 ::
    DUP %0< case :: # 203 ERROROUT ;
    COERCE GRMakeGraph
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME RANGRAPH
::
  CK2&Dispatch
  # 11 ::
    DUP %0< case :: # 203 ERROROUT ;
    DUP %1 %> case :: # 203 ERROROUT ;
    SWAPDUP %0< case :: # 203 ERROROUT ;
    COERCE DUP#0=case :: 2DROP NULL{} ;
    ZERO UNROT { NULLLAM NULLLAM NULLLAM } BIND
    1GETLAM #1+_ONE_DO (DO)
      ZERO 3PUTLAM
      1GETLAM #1+_ONE_DO (DO)
	%RAN 2GETLAM %< IT ::
	  3GETLAM #1+ 3PUTLAM
	  INDEX@ UNCOERCE
	;
      LOOP
      3GETLAM {}N
    LOOP
    1GETABND {}N
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME RANGRAPH2
::
  CK2&Dispatch
  # 11 ::
    DUP %0< case :: # 203 ERROROUT ;
    DUP %1 %> case :: # 203 ERROROUT ;
    SWAPDUP %0< case :: # 203 ERROROUT ;
    COERCE DUP#0=case :: 2DROP NULL{} ;
    ZERO UNROT { NULLLAM NULLLAM NULLLAM } BIND
    1GETLAM #1+_ONE_DO (DO)
      ZERO 3PUTLAM
      INDEX@ #1+_ONE_DO (DO)
	%RAN 2GETLAM %< IT ::
	  3GETLAM #1+ 3PUTLAM
	  INDEX@ UNCOERCE
	;
      LOOP
      3GETLAM {}N
    LOOP
    1GETABND {}N
    ( add reverse edges )
    DUP ' ::
      SWAP UNCOERCE 1LAMBIND ( current node number )
      ' ::
	DUP 3GETLAM %= caseDROP
	3GETLAM GRAddEdge
      ; ListDo
      ABND
    ; ListDo2
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME GETNODE
::
  CK2&Dispatch
  # 51 ::
    OVER LENCOMP UNCOERCE OVER CKNode NOTcase :: # 203 ERROROUT ;
    GRGetNode
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME PUTNODE
::
  CK3&Dispatch
  # 515 ::
    3PICK LENCOMP UNCOERCE 1LAMBIND
    1GETLAM 3PICK CKNode NOTcase :: # 203 ERROROUT ;
    1GETABND OVER GRParseNode NOTcase :: # 203 ERROROUT ;
    GRSetNode
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME ADDNODE
::
  CK2&Dispatch
  # 55 ::
    OVER LENCOMP #1+ UNCOERCE OVER GRParseNode
    NOTcase :: # 203 ERROROUT ;
    >TCOMP
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME ADDNODE2
::
  CK2&Dispatch
  # 55 ::
    OVER LENCOMP #1+ UNCOERCE 1LAMBIND
    1GETLAM OVER GRParseNode NOTcase :: # 203 ERROROUT ;
    SWAPOVER >TCOMP SWAP GRNodeEdges
    ( add reverse edges )
    ' ::
      GRSplitEdge
      OVER 3GETLAM %= case2DROP
      3GETLAM SWAP
      ( graph source target labels )
      GRAddEdgeCK
    ; ListDo
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME DELNODE
::
  CK2&Dispatch
  # 51 ::
    OVER LENCOMP UNCOERCE { NULLLAM NULLLAM } BIND
    1GETLAM 2GETLAM CKNode NOTcase :: # 203 ERROROUT ;
    ( remove node )
    2GETLAM COERCE ListSplitAt CDRCOMP &COMP
    ( for all nodes )
    ' ::
      DUPTYPELIST? NOTcase :: # 203 ERROROUT ;
      GRSplitNode
      ( for all edges of current node )
      ' ::
	DUP GRSplitEdge
	SWAP
	9GETLAM OVER CKNode NOTcase :: # 203 ERROROUT ;
	( edge-info labels node )
	DUP 10GETLAM %= case
	( skip incident edge )
	3DROP
	DUP 10GETLAM %> case
	( update edge )
	:: %1- SWAP GRMakeEdge SWAPDROP ;
	( default case: keep existing edge )
	2DROP
      ; ListMap
      GRMakeNode
    ; ListMap
    ABND
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME ADDEDGE
::
  CK3&Dispatch
  # 515 ::
    3PICK LENCOMP UNCOERCE 1LAMBIND
    1GETLAM 3PICK CKNode NOTcase :: # 203 ERROROUT ;
    1GETABND OVER GRParseEdge NOTcase :: # 203 ERROROUT ;
    GRSplitEdge GRAddEdgeCK
  ;
  # 511 ::
    3PICK LENCOMP UNCOERCE 1LAMBIND
    1GETLAM 3PICK CKNode NOTcase :: # 203 ERROROUT ;
    1GETABND OVER GRParseEdge NOTcase :: # 203 ERROROUT ;
    GRSplitEdge GRAddEdgeCK
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME ADDEDGE2
::
  CK3&Dispatch
  # 515 ::
    3PICK LENCOMP UNCOERCE 1LAMBIND
    1GETLAM 3PICK CKNode NOTcase :: # 203 ERROROUT ;
    1GETABND OVER GRParseEdge NOTcase :: # 203 ERROROUT ;
    GRSplitEdge GRAddEdge2CK
  ;
  # 511 ::
    3PICK LENCOMP UNCOERCE 1LAMBIND
    1GETLAM 3PICK CKNode NOTcase :: # 203 ERROROUT ;
    1GETABND OVER GRParseEdge NOTcase :: # 203 ERROROUT ;
    GRSplitEdge GRAddEdge2CK
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME DELEDGE
::
  CK3&Dispatch
  # 511 ::
    3PICK LENCOMP UNCOERCE 1LAMBIND
    1GETLAM 3PICK CKNode NOTcase :: # 203 ERROROUT ;
    1GETLAM OVER CKNode NOTcase :: # 203 ERROROUT ;
    ABND
    GRDelEdgeCK
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME DELEDGE2
::
  CK3&Dispatch
  # 511 ::
    3PICK LENCOMP UNCOERCE 1LAMBIND
    1GETLAM 3PICK CKNode NOTcase :: # 203 ERROROUT ;
    1GETLAM OVER CKNode NOTcase :: # 203 ERROROUT ;
    ABND
    GRDelEdge2CK
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME NODES
::
  CK1&Dispatch
  #5 GRNodeList
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME REVGRAPH
::
  CK1&Dispatch
  #5 ::
    DUPLENCOMP UNCOERCE 1LAMBIND
    ( copy node infos )
    DUP ' :: GRSplitNode DROP NULL{} GRMakeNode ; ListMap
    SWAP
    ( add reverse edges )
    ' ::
      SWAP UNCOERCE 1LAMBIND ( current node number )
      DUPTYPELIST? NOTcase :: # 203 ERROROUT ;
      GRNodeEdges
      ' ::
	GRSplitEdge
	6GETLAM 3PICK CKNode NOTcase :: # 203 ERROROUT ;
	3GETLAM SWAP GRAddEdge
      ; ListDo
      ABND
    ; ListDo2
    ABND
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME SORTGRAPH
::
  CK1&Dispatch
  #5 ::
    ' ::
      DUPTYPELIST? NOTcase :: # 203 ERROROUT ;
      GRSplitNode
      DUP ' :: GREdgeTarget TYPEREAL? ; ListDoWhile NOTcase :: # 203 ERROROUT ;
      ' :: GREdgeTarget SWAP GREdgeTarget %< ; ListSort
      GRMakeNode
    ; ListMap
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME SUBGRAPH
::
  CK2&Dispatch
  # 55 ::
    ( check node list for validity )
    OVER LENCOMP UNCOERCE OVER GRParseNodeList NOTcase :: # 203 ERROROUT ;
    ( check for duplicates )
    DUP ' %> ListSort %0 SWAP ' :: DUPUNROT %<> ; ListDoWhile SWAPDROP
    NOTcase :: # 203 ERROROUT ;
    ( handle empty graph case )
    OVER NULLCOMP? caseDROP
    ( construct map from old to new node numbers )
    OVER LENCOMP { NULLLAM NULLLAM } BIND
    toLEN_DO (DO)
      2GETLAM INDEX@ UNCOERCE ' %= POSCOMP UNCOERCE
    LOOP
    1GETLAM {}N 2GETLAM 1GETABND UNCOERCE
    ( graph map nodes %nodes )
    ( copy subgraph )
    SWAP 4UNROLL { NULLLAM NULLLAM NULLLAM } BIND
    ' ::
      7GETLAM SWAP GRGetNode
      DUPTYPELIST? NOTcase :: # 203 ERROROUT ;
      GRSplitNode
      ' ::
	GRSplitEdge SWAP
	9GETLAM OVER CKNode NOTcase :: # 203 ERROROUT ;
	COERCE
	10GETLAM SWAP NTHELCOMP NOTcase :: # 203 ERROROUT ;
	DUP %0= case2DROP
	SWAP GRMakeEdge
      ; ListMap
      GRMakeNode
    ; ListMap
    ABND
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME ADDGRAPH
::
  CK2&Dispatch
  # 55 ::
    OVER LENCOMP UNCOERCE
    OVER LENCOMP UNCOERCE
    { NULLLAM NULLLAM } BIND
    ' ::
      DUPTYPELIST? NOTcase :: # 203 ERROROUT ;
      GRSplitNode
      ' ::
	GRSplitEdge
	SWAP
	9GETLAM OVER CKNode NOTcase :: # 203 ERROROUT ;
	10GETLAM %+ SWAP
	GRMakeEdge
      ; ListMap
      GRMakeNode
    ; ListMap
    ABND
    &COMP
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME JOINGRAPH
::
  CK2&Dispatch
  # 55 ::
    OVER LENCOMP UNCOERCE
    OVER LENCOMP UNCOERCE
    2DUP %- UNROT %MAX
    { NULLLAM NULLLAM } BIND
    ( make sure both graphs have the same number of nodes )
    2GETLAM %0> ITE ::
      2GETLAM COERCE
      #1+_ONE_DO (DO)
	NULL{}
      LOOP
      2GETLAM COERCE {}N &COMP
    ;
    ::
      2GETLAM %0= ?SEMI
      SWAP
      2GETLAM %CHS COERCE
      #1+_ONE_DO (DO)
	NULL{}
      LOOP
      2GETLAM %CHS COERCE {}N &COMP
      SWAP
    ;
    1GETABND COERCE 1LAMBIND
    ( merge the two graphs )
    ' ::
      OVER TYPELIST? NOTcase :: # 203 ERROROUT ;
      DUPTYPELIST? NOTcase :: # 203 ERROROUT ;
      GRSplitNode 4ROLL GRSplitNode
      ( point2 labels2 edges2 point1 labels1 edges1 )
      3PICK ' xNOVAL EQUAL ITE ::
	ROTDROP 5ROLL UNROT
      ;
      ::
	6ROLL DROP
      ;
      ( labels2 edges2 point labels1 edges1 )
      4ROLL &COMP
      ( labels2 point labels1 edges1+edges2 )
      SWAP 4ROLL &COMP SWAP
      ( point labels1+labels2 edges1+edges2 )
      GRMakeNode
    ; List2Do
    1GETABND {}N
  ;
;

* iterate over corresponding elements of two lists

NULLNAME List2Do ( list1 list2 prog -> )
::
  1LAMBIND
  #5 SWAP #5
  BEGIN
    ( list1 ofs1 list2 ofs2 )
    2SWAP NEXTCOMPOB ITE ::
      ( list2 ofs2 list1 ofs1' el1 )
      5ROLL 5ROLL
      ( list1 ofs1' el1 list2 ofs2 )
      NEXTCOMPOB ITE ::
        ( list1 ofs1' el1 list2 ofs2' el2 )
        4ROLL SWAP
        ( list1 ofs1' list2 ofs2' el1 el2 )
        TRUE
      ;
      ::
        ( list1 ofs1' el1 list2 )
        4DROP FALSE
      ;
    ;
    ::
      ( list2 ofs2 list1 )
      3DROP FALSE
    ;
  WHILE
    ( list1 ofs1' list2 ofs2' el1 el2 )
    1GETLAM EVAL
    ( list1 ofs1' list2 ofs2' el' )
    5UNROLL
  REPEAT
  ABND
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME NODE\8D
::
  CK1&Dispatch
  #5 GRSplitNode
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME \8DNODE
::
  CK3&Dispatch
  # 255 GRMakeNode
  # 855 ::
    3PICK ' xNOVAL EQUAL NOTcase :: # 202 ERROROUT ;
    GRMakeNode
  ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME ADJ
::
  CK1&Dispatch
  #5 GRNodeAdj
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME EDGES
::
  CK1&Dispatch
  #5 GRNodeEdges
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME POINT
::
  CK1&Dispatch
  #5 GRNodePoint
;

NULLNAME Node?
::
  DUPNULLCOMP? casedrop TRUE
  ( take a look at the header of the list )
  DUP CARCOMP
  DUPTYPECMP? case2drop TRUE ( must be a node )
  DUPTYPETAG? case2drop TRUE ( must be a node )
  DUPTYPEREAL? casedrop ::
    CDRCOMP DUPNULLCOMP? casedrop TRUE ( FIXME we can't be sure here )
    CARCOMP
    DUPTYPEREAL? casedrop TRUE ( must be a node )
    DUPTYPELIST? casedrop TRUE ( must be a node )
    DUPTYPETAG? casedrop FALSE ( must be an edge )
    # 203 ERROROUT ( malformed list )
  ;
  DUPTYPELIST? case2drop TRUE ( must be a node )
  # 203 ERROROUT ( malformed list )
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME LABELS
::
  CK1&Dispatch
  #5 :: DUP Node? ITE GRNodeLabels GREdgeLabels ;
  #1 GREdgeLabels
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME EDGE\8D
::
  CK1&Dispatch
  #1 GRSplitEdge
  #5 GRSplitEdge
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME \8DEDGE
::
  CK2&Dispatch
  # 15 GRMakeEdge
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME TARGET
::
  CK1&Dispatch
  #1 NOP
  #5 CARCOMP
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME GETLABEL
::
  CK2&Dispatch
  # 13 :: 2DROP ' xNOVAL ;
  # 16 :: 2DROP ' xNOVAL ;
  # 17 :: 2DROP ' xNOVAL ;
  # 53 :: OVER Node? ITE GRNodeLabel GREdgeLabel ;
  # 56 :: OVER Node? ITE GRNodeLabel GREdgeLabel ;
  # 57 :: OVER Node? ITE GRNodeLabel GREdgeLabel ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME PUTLABEL
::
  CK3&Dispatch
  # 130 GRSetEdgeLabel
  # 160 GRSetEdgeLabel
  # 170 GRSetEdgeLabel
  # 530 :: 3PICK Node? ITE GRSetNodeLabel GRSetEdgeLabel ;
  # 560 :: 3PICK Node? ITE GRSetNodeLabel GRSetEdgeLabel ;
  # 570 :: 3PICK Node? ITE GRSetNodeLabel GRSetEdgeLabel ;
;

ASSEMBLE
        CON(1)        8                * Tell parser 'Non algebraic'
RPL
xNAME DELLABEL
::
  CK2&Dispatch
  # 13 DROP
  # 16 DROP
  # 17 DROP
  # 53 :: OVER Node? ITE GRDelNodeLabel GRDelEdgeLabel ;
  # 56 :: OVER Node? ITE GRDelNodeLabel GRDelEdgeLabel ;
  # 57 :: OVER Node? ITE GRDelNodeLabel GRDelEdgeLabel ;
;
