RPL ( Ensure RPLCOMP and the syntax highlighter are in RPL mode )
( D:\Dokumente und Einstellungen\Administrator\Eigene Dateien\HP\Projects\GRW\util.s, part of the GRW.hpp project, created by <> on 17.12.2001 )

INCLUDE GRW.H

* generic list mapping functions *********************************************

* These are all implemented employing the usual return stack trickery, and
* hence are a lot more efficient than loops using NTHCOMP/PUTLIST.

* For each of these operations, the list and prog arguments are taken from the
* stack, and then the prog is applied to each list element in turn. The mapped
* program may access and consume additional args above the list and prog
* arguments. For instance, folding a binary operation over a list starting
* from an initial value can be done as follows: %0 { %1 %2 %3 } ' %+ ListDo.

* CAVEAT: Each of these routines uses a temporary environment, which makes it
* difficult to access NULLLAMs in outer environments, unless you know
* beforehand the ListDo level at which the given program will be executed. The
* number of lambda variables of each routine is specified below.

* ListDo: ( list prog -> ob ... )
* applies prog to each member of the list, leaving the results on the stack
* uses one temporary

NULLNAME ListDo
::
  OVER NULLCOMP? case2DROP
  1LAMBIND
  >R 'R
  [BEGIN]
    1GETLAM EVAL
    RSWAP ticR
  [WHILE]
    RSWAP
  [REPEAT]
  ABND
;

* ListDo2: ( list prog -> ob ... )
* like ListDo, but prog also gets the index of the current element in level 2
* uses two temporaries

NULLNAME ListDo2
::
  OVER NULLCOMP? case2DROP
  #0 { NULLLAM NULLLAM } BIND
  >R 'R
  [BEGIN]
    1GETLAM #1+ DUP 1PUTLAM SWAP
    2GETLAM EVAL
    RSWAP ticR
  [WHILE]
    RSWAP
  [REPEAT]
  ABND
;

* ListDoWhile: ( list prog -> ob ... T|F )
* ListDo with result checking
* uses two temporaries

* Here, prog is expected to return a flag in level 1, and the routine keeps
* iterating only as long as the returned flag is TRUE; the outcome of the last
* iteration (TRUE or FALSE) is returned on level 1, the remaining results are
* placed on the stack above level 1.

NULLNAME ListDoWhile
::
  OVER NULLCOMP? case2drop TRUE
  TRUE { NULLLAM NULLLAM } BIND
  >R 'R
  [BEGIN]
    2GETLAM EVAL
    RSWAP ITE ticR :: FALSE 1PUTLAM RSWAP RDROP FALSE ;
  [WHILE]
    RSWAP
  [REPEAT]
  1GETABND
;

* ListDo2While: ( list prog -> ob ... T|F )
* ListDo2 with result checking
* uses three temporaries

NULLNAME ListDo2While
::
  OVER NULLCOMP? case2drop TRUE
  #0 TRUE { NULLLAM NULLLAM NULLLAM } BIND
  >R 'R
  [BEGIN]
    2GETLAM #1+ DUP 2PUTLAM SWAP
    3GETLAM EVAL
    RSWAP ITE ticR :: FALSE 1PUTLAM RSWAP RDROP FALSE ;
  [WHILE]
    RSWAP
  [REPEAT]
  1GETABND
;

* The following routines are wrappers around the ListDo operations which
* collect the results in a list. They use an extra local environment
* containing one temporary, in addition to the environment of the
* corresponding ListDo function.

* The mapped function may access extra args above the list and prog arguments,
* but must not consume these arguments. This is because the length of the
* computed list is calculated by comparing the stack depths before and after
* the execution of the ListDo function.

* ListMap: ( list prog -> list' )

NULLNAME ListMap
::
  DEPTH #2- 1LAMBIND
  ListDo
  DEPTH 1GETABND #- {}N
;

* ListMap2: ( list prog -> list' )

NULLNAME ListMap2
::
  DEPTH #2- 1LAMBIND
  ListDo2
  DEPTH 1GETABND #- {}N
;

* ListMapWhile: ( list prog -> list' T|F )

NULLNAME ListMapWhile
::
  DEPTH #2- 1LAMBIND
  ListDoWhile
  DEPTH #1- 1GETABND #- SWAP OVER#2+UNROL {}N SWAP
;

* ListMap2While: ( list prog -> list' T|F )

NULLNAME ListMap2While
::
  DEPTH #2- 1LAMBIND
  ListDo2While
  DEPTH #1- 1GETABND #- SWAP OVER#2+UNROL {}N SWAP
;

* min and max of a list *****************************************************

* ListMin: ( list prog -> ob )
* find the element of the given list which, when applying the given prog,
* gives the smallest real value; returns {} if list is empty

NULLNAME ListMin
::
  OVER NULLCOMP? caseDROP
  %MAXREAL CRUNCH NULL{} { NULLLAM NULLLAM NULLLAM } BIND
  >R 'R
  [BEGIN]
    DUP 3GETLAM EVAL
    DUP 2GETLAM %< ITE :: 2PUTLAM 1PUTLAM ; 2DROP
    RSWAP ticR
  [WHILE]
    RSWAP
  [REPEAT]
  1GETABND
;

* ListMax: ( list prog -> ob )
* find the element of the given list which, when applying the given prog,
* gives the smallest real value; returns {} if list is empty

NULLNAME ListMax
::
  OVER NULLCOMP? caseDROP
  %-MAXREAL CRUNCH NULL{} { NULLLAM NULLLAM NULLLAM } BIND
  >R 'R
  [BEGIN]
    DUP 3GETLAM EVAL
    DUP 2GETLAM %> ITE :: 2PUTLAM 1PUTLAM ; 2DROP
    RSWAP ticR
  [WHILE]
    RSWAP
  [REPEAT]
  1GETABND
;

* list splitting ************************************************************

* ListSplit: ( list pred -> list1 list2 )
* split a list into the longest prefix of elements satisfying pred
* (returned on level 2), and the remainder of the list (level 1)

NULLNAME ListSplit
::
  OVER NULLCOMP? casedrop DUP
  #1 { NULLLAM NULLLAM } BIND
  DUP >R 'R
  [BEGIN]
    2GETLAM EVAL
    DUP IT :: 1GETLAM #1+ 1PUTLAM ;
    RSWAP ITE ticR :: RSWAP RDROP FALSE ;
  [WHILE]
    RSWAP
  [REPEAT]
  1GETABND ListSplitAt
;

* ListSplitAt: ( list # -> list1 list2 )
* split a list at the specified index (>= 1, <= size(list)+1)

NULLNAME ListSplitAt
::
  #1- 1LAMBIND
  INNERCOMP DUP #1+ SWAP 1GETABND #- SWAP
  metatail {}N OVER#2+UNROL {}N SWAP
;

* sorting *******************************************************************

* This is a generic version of Joseph K. Horn's binary insertion sort
* algorithm, shamelessly pilfered from Goodies Disk #8. This algorithm does
* stable sorting.

* ListSort: ( list ord -> list' )
* sort a list (in ascending order) w.r.t. the given order predicate; ord is
* assumed to return TRUE iff the list element on level 2 is greater than that
* on level 1

NULLNAME ListSort
::
  SWAPINCOMP get1
  { NULLLAM NULLLAM } BIND
  2GETLAM
  #1+_ONE_DO (DO)
    2GETLAM ROLL #0 INDEX@
    BEGIN 2DUP SWAP#- #2/ DUP#0<>
    WHILE
      OVERSWAP #-DUP #4+PICK 5PICK 1GETLAM EVAL
      ITE ROTDROPSWAP SWAPDROP
    REPEAT
    ROT2DROP UNROLL
  LOOP
  2GETLAM ABND {}N
;

* ListInsert: ( list ob ord -> list' )
* The binary insertion algorithm from the above routine; used to insert a
* new element into an already sorted list.

NULLNAME ListInsert
::
  1LAMBIND SWAPINCOMP get1 SWAP #1+ 1GETABND
  { NULLLAM NULLLAM } BIND
  #0 2GETLAM
  BEGIN 2DUP SWAP#- #2/ DUP#0<>
  WHILE
    OVERSWAP #-DUP #4+PICK 5PICK 1GETLAM EVAL
    ITE ROTDROPSWAP SWAPDROP
  REPEAT
  ROT2DROP UNROLL
  2GETLAM ABND {}N
;
